/*
 * This is using the distalg/messaging stuff.
 * Hopefully this will be made a little easier by that.
*/

module gnutella

open std/ord
open gnutella/messaging2

sig Number {
	prev: option Number,
	next: option Number
} {
	prev = OrdPrev(this)
	next = OrdNext(this)
}

// represents a node in the network
// gnutella calls them servants because they
// are both clients and servers
sig Servant extends Node {
	routingTable: Tick->Msg->Node,
	connection: set Servant
//	vNewConnection: Tick->Node
}

static sig Pinger extends Node {
}
/*
sig ServantState extends NodeState {
	routingTable: Msg->Node,
	vNewConnection: Node
}*/

fun Visible(n2: Node) {
  let num = Ord[Number].first | {
    n2 != Pinger && { 
	(one num && n2 in Pinger.connection) ||
	one OrdNext(num) && n2 in Pinger.connection.connection ||
	one OrdNext(OrdNext(num)) && n2 in Pinger.connection.connection.connection ||
	one OrdNext(OrdNext(OrdNext(num))) && n2 in Pinger.connection.connection.connection.connection ||
	one OrdNext(OrdNext(OrdNext(OrdNext(num)))) && n2 in Pinger.connection.connection.connection.connection.connection 
    }
  }
}

/*fact OnlyAddNewConnections{
	all n: Node|  {
		all t:Tick |
			n.vNewConnection[OrdPrev(t)] in n.vNewConnection[t]
		no n.vNewConnection[Ord[Tick].first] 
	}	
	//some Node.vNewConnection
}*/

sig GnuMsg extends Msg {
	TTL: Number,
	GUID: GnuMsg,
	servant: Node
}{
	some to
}

disj sig Ping,Pong extends GnuMsg{}

fact {
	GnuMsg = Msg
}

fact firstTickEmpty{
	all n: Node |
		no n.routingTable[Ord[Tick].first]
}

fun NoPhantomEntriesInTable(n: Node, t:Tick) {
	let msgs =  Ping & OrdPrevs(t).read[n] | {
		no msgs => no(n.routingTable[t]),
		let k = msgs->msgs & iden[Msg] |
 		(n.routingTable[t]) = (~(k.(Msg$GUID))).(k.(Msg$from))
	}
	
}

fact infoAboutGuids{
	Msg.GUID in Ping 
}

fun SendFirstPing(self: Node, sees, reads, sends, needsToSend: set Msg) {
	one sends
	sends.TTL = Ord[Number].last
	#needsToSend = 1
	sends.to = self.connection
	sends in Ping
	sends.GUID = sends
}

/* Problem when two messages in the set have that GUID
   Because both arrived from Different people 
   The message that fails has a different to than the whole thing
*/
fun isOurForwardedPing(self: Node, reads: Msg, send:Msg) {
	one reads
	send in Ping
	send.GUID = reads.GUID
//	send.TTL != Ord[Number].first 
	send.to = self.connection - reads.from
	send.TTL = reads.TTL.prev 
}

fun isForwardedPong(self: Node, reads: Msg, send: Msg, table: Msg->Node) {
	one reads
	reads in Pong
	send.to = (reads.GUID).table
	send.GUID = reads.GUID
	send in Pong
	send.servant = reads.servant
}

fun isOurPing(self:Node, reads: Msg, send: Msg) {
	send in Ping
	send.GUID = send
	send.from = Pinger
	send.to = self.connection
	send.TTL = Ord[Number].last
}

fun isReplyPong(self: Node, reads: Msg, send: Msg) {
	reads in Ping
	send in Pong
	one send.to
	send.to = reads.from
	send.GUID = reads.GUID
	send.servant = self
}

fun PingDeath(read: Msg, table: Msg->Node) {
	read.TTL = Ord[Number].first || no (read.to.connection - read.from) || some (read.GUID).table
}

fun PongDeath(read: Msg, table: Msg->Node) {
	no (read.GUID).table
}

/*
fun addToVConnection(read: Msg, self: Node, tick:Tick) {
	read.servant in self.vNewConnection[OrdNext(tick)]
}

fun NoPhantomVConnections(self: Node, tick:Tick) {
	all n: self.vNewConnection[tick] |
		some msg: OrdPrevs(tick).read[self] & Pong|
			msg.servant = n
}*/

fun ConnectedInTheEnd(pinger:Node, other:Node) {
	other in (Tick.read[pinger] & Pong).servant
}


fun addToRoutingTable(received: Msg, send: Msg, tick:Tick) {
	no OrdNext(tick) ||
	received.GUID->received.from in send.from.routingTable[OrdNext(tick)]
}

fun Transition(self: Node, sees, reads, sends, needsToSend: set Msg, tick: Tick) {
	reads = sees
	all weSend : sends | { 
			isOurForwardedPing(self, reads, weSend) || isOurPing(self, reads, weSend) 
 			|| isReplyPong(self, reads, weSend) || isForwardedPong(self, reads, weSend, self.routingTable[tick])
	}
	one reads => {
	reads in Ping =>  {
		PingDeath(reads, self.routingTable[tick]) || 
		one msg: sends | {
			isOurForwardedPing(self, reads, msg) 
			addToRoutingTable(reads, msg, tick)				
		}
		one msg: sends | 
			isReplyPong(self,reads, msg)
	}, reads in Pong => {
		PongDeath(reads, self.routingTable[tick]) || 
			one msg: sends | 
				isForwardedPong(self,reads, msg, self.routingTable[tick])
	  }
	}

	NoPhantomEntriesInTable(self, tick)
	#needsToSend = #sends	
}
	


fun Transitions() {
	SendFirstPing(Pinger, 
			      
				Ord[Tick].first.visible[Pinger],
				Ord[Tick].first.read[Pinger], Ord[Tick].first.sent[Pinger], 
				Ord[Tick].first.needsToSend[Pinger])
	all n:Node | {
		all t: Tick - Ord[Tick].last | 
			Transition(n, t.visible[n], 
					t.read[n], t.sent[n], t.needsToSend[n], t)
		Transition(n, Ord[Tick].last.visible[n], Ord[Tick].last.read[n], 
				Ord[Tick].last.sent[n], Ord[Tick].last.needsToSend[n], Ord[Tick].last)
	}

}



fact MsgsOnlyThroughConnections{
	all m:Msg |
		no m.to || m.to in m.from.connection
}

fact doubleconnect {
	Node$connection = ~Node$connection
}

fact NoLoops{
	no (iden[Node] & Node$connection)
}

fact AllNodesUsed{
	Node = univ[Node]
}

fact ConnectAll{
	all n:Node |
		Node in n.^connection 
}

fun LotsConnections() {
	all n:Node |
		#n.connection =< 3
}

fun InALine() {
	all n:Node |
		#n.connection < 3
}

fun LeaderNotFullyConnected() {
	Pinger.connection + Pinger != Node
	
}

fun LeaderTwoConnections() {
	#Pinger.connection > 1
}

fun LeaderOneConnection() {
	#Pinger.connection = 1
}

fun SomeNodeSplits() {
	some n:Node |
		#n.connection > 2
}

fun AllConnected() {
	all n: Node |
		#n.connection = 3
}

fun Ring() {
	all n:Node |
		#n.connection = 2
}

fact DeliveredOnNextTick {
  all m: Msg | 
    (m.sentOn != Ord[Tick].last) => all n: m.to | 
        one OrdNext(m.sentOn).visible[n]
}
fun Show() {
	ReadInOrder()
	NoLostMessages()
	NoMessageShortage()
	Transitions()
}

fun ShowInARing() {
	Show()
	Ring()
}

fun SeparateRing() {
	all n:Node-Pinger |
		#n.connection > 1
}

fun ShowInALine() {
	Show()
	//LeaderTwoConnections()
	LeaderOneConnection()
	InALine()
}

fun ShowVisible() {
	SeparateRing()
//	LeaderNotFullyConnected()
	//InALine()
	LeaderOneConnection()
}

fun TestConnection() {
	Show()
//	LeaderNotFullyConnected()
	LeaderOneConnection()
	SeparateRing()
}

assert AllConnects {
  (TestConnection()) => 
    all n2: Node | 
		ConnectedInTheEnd(Pinger, n2) <=>
			n2 in Node
	
}

assert AllConnectsTTL {
  (TestConnection()) =>
    all n2: Node | 
		ConnectedInTheEnd(Pinger, n2) <=>
			Visible(n2)
}

run ShowVisible for 1 Number,  4 Node, 1 NodeState, 2 Tick, 2 Msg

// These check for 3 nodes, TTL=1
run TestConnection for 1 Number,  3 Node, 1 NodeState, 5 Tick, 4 Msg
check AllConnects for 1 Number,  3 Node, 1 NodeState, 5 Tick, 4 Msg
check AllConnectsTTL for 1 Number,  3 Node, 1 NodeState, 5 Tick, 4 Msg

// 3 node, TTL=2
run TestConnection for 2 Number,  3 Node, 1 NodeState, 5 Tick, 5 Msg
check AllConnects for 2 Number,  3 Node, 1 NodeState, 5 Tick, 5 Msg
check AllConnectsTTL for 2 Number,  3 Node, 1 NodeState, 5 Tick, 5 Msg

// 4 Node, TTL=2, 5 Tick/Msg
run TestConnection for 2 Number, 4 Node, 1 NodeState, 5 Tick, 5 Msg
check AllConnects for 2 Number, 4 Node, 1 NodeState, 5 Tick, 5 Msg
check AllConnectsTTL for 2 Number, 4 Node, 1 NodeState, 5 Tick, 5 Msg

// These check for 4 nodes TTL=2
run TestConnection for 2 Number, 4 Node, 1 NodeState, 7 Tick, 7 Msg
check AllConnects for 2 Number, 4 Node, 1 NodeState, 7 Tick, 7 Msg
check AllConnectsTTL for 2 Number, 4 Node, 1 NodeState, 7 Tick, 7 Msg

